
class InstallError < StandardError; end


class Installer

  Version   = '3.1.4'
  Copyright = 'Copyright (c) 2000-2003 Minero Aoki'


  @toplevel = nil

  def self.declare_toplevel_installer( inst )
    raise ArgumentError, 'two toplevel installers declared' if @toplevel
    @toplevel = inst
  end

  def self.toplevel_installer
    @toplevel
  end


  FILETYPES = %w( bin lib ext data )

  include FileOperations

  def initialize( config, opt, srcroot, objroot )
    @config = config
    @options = opt
    @srcdir = File.expand_path(srcroot)
    @objdir = File.expand_path(objroot)
    @currdir = '.'
  end

  def inspect
    "#<#{self.class} #{__id__}>"
  end

  #
  # configs/options
  #

  def get_config( key )
    @config[key]
  end

  alias config get_config

  def set_config( key, val )
    @config[key] = val
  end

  def no_harm?
    @options['no-harm']
  end

  def verbose?
    @options['verbose']
  end

  def verbose_off
    save, @options['verbose'] = @options['verbose'], false
    yield
    @options['verbose'] = save
  end

  #
  # srcdir/objdir
  #

  attr_reader :srcdir
  alias srcdir_root srcdir
  alias package_root srcdir

  def curr_srcdir
    "#{@srcdir}/#{@currdir}"
  end

  attr_reader :objdir
  alias objdir_root objdir

  def curr_objdir
    "#{@objdir}/#{@currdir}"
  end

  def srcfile( path )
    curr_srcdir + '/' + path
  end

  def srcexist?( path )
    File.exist? srcfile(path)
  end

  def srcdirectory?( path )
    dir? srcfile(path)
  end
  
  def srcfile?( path )
    FileTest.file? srcfile(path)
  end

  def srcentries( path = '.' )
    Dir.open(curr_srcdir + '/' + path) {|d|
        return d.to_a - %w(. ..) - hookfilenames
    }
  end

  def srcfiles( path = '.' )
    srcentries(path).select {|fname|
        FileTest.file? File.join(curr_srcdir, path, fname)
    }
  end

  def srcdirectories( path = '.' )
    srcentries(path).select {|fname|
        dir? File.join(curr_srcdir, path, fname)
    }
  end

  def dive_into( rel )
    return unless dir?("#{@srcdir}/#{rel}")

    dir = File.basename(rel)
    Dir.mkdir dir unless dir?(dir)
    prevdir = Dir.pwd
    Dir.chdir dir
    $stderr.puts '---> ' + rel if verbose?
    @currdir = rel
    yield
    Dir.chdir prevdir
    $stderr.puts '<--- ' + rel if verbose?
    @currdir = File.dirname(rel)
  end

  #
  # TASK config
  #

  def exec_config
    exec_task_traverse 'config'
  end

  def config_dir_bin( rel )
  end

  def config_dir_lib( rel )
  end

  def config_dir_ext( rel )
    extconf if extdir? curr_srcdir
  end

  def extconf
    opt = @options['config-opt'].join(' ')
    command "#{config('ruby-prog')} #{curr_srcdir}/extconf.rb #{opt}"
  end

  def config_dir_data( rel )
  end

  #
  # TASK setup
  #

  def exec_setup
    exec_task_traverse 'setup'
  end

  def setup_dir_bin( relpath )
    all_files_in(curr_srcdir()).each do |fname|
      add_rubypath "#{curr_srcdir}/#{fname}"
    end
  end

  SHEBANG_RE = /\A\#!\s*\S*ruby\S*/

  def add_rubypath( path )
    $stderr.puts %Q<set #! line to "\#!#{config('ruby-path')}" for #{path} ...> if verbose?
    return if no_harm?

    tmpfile = File.basename(path) + '.tmp'
    begin
      File.open(path) {|r|
      File.open(tmpfile, 'w') {|w|
          first = r.gets
          return unless SHEBANG_RE === first   # reject '/usr/bin/env ruby'

          w.print first.sub(SHEBANG_RE, '#!' + config('ruby-path'))
          w.write r.read
      } }
      mv tmpfile, File.basename(path)
    ensure
      rm_f tmpfile if File.exist? tmpfile
    end
  end

  def setup_dir_lib( relpath )
  end

  def setup_dir_ext( relpath )
    make if extdir?(curr_srcdir)
  end

  def setup_dir_data( relpath )
  end

  #
  # TASK install
  #

  def exec_install
    exec_task_traverse 'install'
  end

  def install_dir_bin( rel )
    install_files target_filenames(), config('bin-dir') + '/' + rel, 0755
  end

  def install_dir_lib( rel )
    install_files target_filenames(), config('rb-dir') + '/' + rel, 0644
  end

  def install_dir_ext( rel )
    install_dir_ext_main File.dirname(rel) if extdir?(curr_srcdir)
  end

  def install_dir_ext_main( rel )
    install_files allext('.'), config('so-dir') + '/' + rel, 0555
  end

  def install_dir_data( rel )
    install_files target_filenames(), config('data-dir') + '/' + rel, 0644
  end

  def install_files( list, dest, mode )
    mkdir_p dest, @options['install-prefix']
    list.each do |fname|
      install fname, dest, mode, @options['install-prefix']
    end
  end

  def target_filenames
    if FileTest.file? "#{curr_srcdir()}/MANIFEST"
      mapdir(target_filenames_MANIFEST())
    else
      mapdir(target_filenames_AUTO())
    end
  end

  def mapdir( filelist )
    filelist.map {|fname|
        if File.exist? fname   # current objdir == '.'
          fname
        else
          File.join(curr_srcdir(), fname)
        end
    }
  end

  def target_filenames_MANIFEST
    File.read_all("#{curr_srcdir()}/MANIFEST").split
  end
  
  # picked up many entries from cvs-1.11.1/src/ignore.c
  REJECT_PATTERNS = %w( 
      core RCSLOG tags TAGS .make.state
      .nse_depinfo #* .#* cvslog.* ,* .del-* *.a *.olb *.o *.obj
      *.so *.Z *~ *.old *.elc *.ln *.bak *.BAK *.orig *.rej *.exe _$* *$

      *.org *.in .*
  ).map {|pattern|
      Regexp.compile('\A' + pattern.gsub(/[\.\$]/) {|s| '\\' + s }.gsub(/\*/, '.*') + '\z')
  }

  def target_filenames_AUTO
    (existfiles() - hookfiles()).reject {|fname|
        REJECT_PATTERNS.any? {|re| re === fname }
    }
  end

  def existfiles
    all_files_in(curr_srcdir()) | all_files_in(curr_objdir())
  end

  def hookfiles
    %w( pre-%s post-%s pre-%s.rb post-%s.rb ).map {|fmt|
        %w( config setup install clean ).map {|t| sprintf(fmt, t) }
    }.flatten
  end

  def allext( dir )
    _allext(dir) or raise InstallError,
        "no extention exists: Have you done 'ruby #{$0} setup' ?"
  end

  DLEXT = /\.#{ ::Config::CONFIG['DLEXT'] }\z/

  def _allext( dir )
    Dir.open(dir) {|d|
        return d.select {|fname| DLEXT === fname }
    }
  end

  #
  # TASK clean
  #

  def exec_clean
    exec_task_traverse 'clean'
    rm_f 'config.save'
    rm_f 'InstalledFiles'
  end

  def clean_dir_bin( rel )
  end

  def clean_dir_lib( rel )
  end

  def clean_dir_ext( rel )
    make 'clean' if FileTest.file?('Makefile')
  end

  def clean_dir_data( rel )
  end

  #
  # TASK distclean
  #

  def exec_distclean
    exec_task_traverse 'distclean'
    rm_f 'config.save'
    rm_f 'InstalledFiles'
  end

  def distclean_dir_bin( rel )
  end

  def distclean_dir_lib( rel )
  end

  def distclean_dir_ext( rel )
    make 'distclean' if FileTest.file?('Makefile')
  end

  #
  # lib
  #
  
  def make( task = '' )
    command config('make-prog') + ' ' + task
  end

  def exec_task_traverse( task )
    run_hook 'pre-' + task
    FILETYPES.each do |type|
      if config('without-ext') == 'yes' and type == 'ext'
        $stderr.puts 'skipping ext/* by user option' if verbose?
        next
      end
      traverse task, type, task + '_dir_' + type
    end
    run_hook 'post-' + task
  end

  def traverse( task, rel, mid )
    dive_into(rel) {
        run_hook 'pre-' + task
        __send__ mid, rel.sub(%r[\A.*?(?:/|\z)], '')
        all_dirs_in(curr_srcdir()).each do |d|
          traverse task, rel + '/' + d, mid
        end
        run_hook 'post-' + task
    }
  end

  def run_hook( name )
    try_run_hook curr_srcdir + '/' + name           or
    try_run_hook curr_srcdir + '/' + name + '.rb'
  end

  def try_run_hook( fname )
    return false unless FileTest.file?(fname)

    env = self.dup
    begin
      env.instance_eval File.read_all(fname), fname, 1
    rescue
      raise InstallError, "hook #{fname} failed:\n" + $!.message
    end
    true
  end

  def extdir?( dir )
    File.exist? dir + '/MANIFEST'
  end

end

